import requests
import logging
from typing import List
from ..utils.timer import Timer
from ..utils.transformations import parameters_to_fews

LOGGER = logging.getLogger(__name__)


def get_filters(
    url: str,
    filter_id: str = None,
    document_format: str = "PI_JSON",
    verify: bool = False,
    logger=LOGGER,
) -> List[dict]:
    """
    Get FEWS qualifiers as a pandas DataFrame

    Args:
        url (str): url Delft-FEWS PI REST WebService.
        E.g. http://localhost:8080/FewsWebServices/rest/fewspiservice/v1/filters
        filter_id (str): the FEWS id of the filter to pass as request parameter
        document_format (str): request document format to return. Defaults to PI_JSON.
        verify (bool, optional): passed to requests.get verify parameter.
        Defaults to False.
        logger (logging.Logger, optional): Logger to pass logging to. By
        default, a logger will ge created.

    Returns:
        df (pandas.DataFrame): Pandas dataframe with index "id" and columns
        "name" and "group_id".

    """

    # do the request
    timer = Timer(logger)
    parameters = parameters_to_fews(locals())
    response = requests.get(url, parameters, verify=verify)
    timer.report("Filters request")

    # parse the response
    result = []
    if response.status_code == 200:
        if "filters" in response.json().keys():
            result = response.json()["filters"]
        timer.report("Filters parsed")
    else:
        logger.error(f"FEWS Server responds {response.text}")

    return result



import requests
from xml.etree import ElementTree
import pandas as pd
import logging
from ..utils.timer import Timer

NS = "{http://www.wldelft.nl/fews/PI}"
LOGGER = logging.getLogger(__name__)
COLUMNS = ["id", "name", "group_id"]


def _element_to_tuple(qualifier_element: ElementTree.Element) -> tuple:
    """
    Parses a qualifier element to a tuple

    Args:
        qualifier_element (xml.etree.ElementTree.Element): ET.Element with
        Delft-FEWS qualifier tags.

    Returns:
        tuple: qualifier properties (id, name, group_id). If not present they
        will be None.

    """

    def __get_text(element):
        if element is not None:
            return element.text
        else:
            return element

    ident = qualifier_element.get("id")
    name = __get_text(next(qualifier_element.iter(tag=f"{NS}name"), None))
    group_id = __get_text(next(qualifier_element.iter(tag=f"{NS}groupId"), None))

    return (ident, name, group_id)


def get_qualifiers(url: str, verify: bool = False, logger=LOGGER) -> pd.DataFrame:
    """
    Get FEWS qualifiers as Pandas DataFrame

    Args:
        url (str): url Delft-FEWS PI REST WebService.
        E.g. http://localhost:8080/FewsWebServices/rest/fewspiservice/v1/qualifiers
        verify (bool, optional): passed to requests.get verify parameter.
        Defaults to False.
        logger (logging.Logger, optional): Logger to pass logging to. By
        default, a new logger will ge created.

    Returns:
        df (pandas.DataFrame): Pandas dataframe with index "id" and columns
        "name" and "group_id".

    """

    # do the request
    timer = Timer(logger)
    response = requests.get(url, verify=False)
    timer.report("Qualifiers request")

    # parse the response
    if response.status_code == 200:
        tree = ElementTree.fromstring(response.content)
        qualifiers_tree = [i for i in tree.iter(tag=f"{NS}qualifier")]
        qualifiers_tuple = (_element_to_tuple(i) for i in qualifiers_tree)
        df = pd.DataFrame(qualifiers_tuple, columns=COLUMNS)
        timer.report("Qualifiers parsed")
    else:
        logger.error(f"FEWS Server responds {response.text}")
        df = pd.DataFrame(columns=COLUMNS)
    df.set_index("id", inplace=True)

    return df



import requests
import pandas as pd
import logging
from ..utils.timer import Timer
from ..utils.transformations import parameters_to_fews
from typing import List, Union
from ..time_series import TimeSeriesSet
from datetime import datetime
from fewspy.io.read_xml import read_xml_from_string
from fewspy.io.read_netcdf import read_netcdf_from_content


LOGGER = logging.getLogger(__name__)


def _ts_or_headers(only_headers=False):
    if only_headers:
        return "Headers {status}"
    else:
        return "TimeSeries {status}"


def get_time_series(
    url: str,
    filter_id: str,
    location_ids: Union[str, List[str]] = None,
    parameter_ids: Union[str, List[str]] = None,
    qualifier_ids: Union[str, List[str]] = None,
    start_time: datetime = None,
    end_time: datetime = None,
    thinning: int = None,
    only_headers: bool = False,
    omit_missing: bool = True,
    show_statistics: bool = False,
    document_format: str = "PI_JSON",
    verify: bool = False,
    logger=LOGGER,
) -> pd.DataFrame:
    """
    Get FEWS qualifiers as a pandas DataFrame

    Args:
        url (str): url Delft-FEWS PI REST WebService.
        E.g. http://localhost:8080/FewsWebServices/rest/fewspiservice/v1/qualifiers
        filter_id (str): the FEWS id of the filter to pass as request parameter
        location_ids (list): list with FEWS location ids to extract timeseries from. Defaults to None.
        parameter_ids (list): list with FEWS parameter ids to extract timeseries from. Defaults to None.
        qualifier_ids (list): list with FEWS qualifier ids to extract timeseries from. Defaults to None.
        start_time (datetime.datetime): datetime-object with start datetime to use in request. Defaults to None.
        end_time (datetime.datetime): datetime-object with end datetime to use in request. Defaults to None.
        thinning (int): integer value for thinning parameter to use in request. Defaults to None.
        only_headers (bool): if True, only headers will be returned. Defaults to False.
        omit_missing (bool): if True, no missings values will be returned. Defaults to True.
        show_statistics (bool): if True, time series statistics will be included in header. Defaults to False.
        document_format (str): request document format to return. Defaults to PI_JSON.
        verify (bool, optional): passed to requests.get verify parameter.
        Defaults to False.
        logger (logging.Logger, optional): Logger to pass logging to. By
        default, a logger will ge created.

    Returns:
        df (pandas.DataFrame): Pandas dataframe with index "id" and columns
        "name" and "group_id".

    """
    report_string = _ts_or_headers(only_headers)

    # do the request
    timer = Timer(logger)
    parameters = parameters_to_fews(locals())
    response = requests.get(url, parameters, verify=verify)
    timer.report(report_string.format(status="request"))

    # parse the response
    if response.ok:
        logger.debug(response.url)
        if document_format == "PI_JSON":
            pi_time_series = response.json()
            time_series_set = TimeSeriesSet.from_dict(pi_time_series)
        elif document_format == "PI_XML":
            time_series_set = read_xml_from_string(response.text)
        elif document_format == "PI_NETCDF":
            time_series_set = read_netcdf_from_content(response.content)
        timer.report(report_string.format(status="parsed"))
        if time_series_set.empty:
            logger.debug(f"FEWS WebService request passing empty set: {response.url}")
    else:
        logger.error(f"FEWS WebService request {response.url} responds {response.text}")
        time_series_set = TimeSeriesSet()

    return time_series_set



from pathlib import Path


def get_header_file(data_file: Path) -> Path:
    """Path to header-file based on data-file

    Args:
        data_file (Path): Path to data-file

    Returns:
        Path: Path to header-file
    """
    return data_file.with_name(f"{data_file.stem}_header{data_file.suffix}")



from datetime import datetime
from config import api


LOCATION_IDS = ["NL34.HL.KGM156.HWZ1", "NL34.HL.KGM156.LWZ1"]
PARAMETER_IDS = ["Q [m3/s] [NVT] [OW]", "WATHTE [m] [NAP] [OW]"]
QUALIFIER_IDS = ["productie"]

time_series_set = api.get_time_series(
    filter_id="WDB_OW_KGM",
    location_ids=LOCATION_IDS,
    start_time=datetime(2022, 5, 1),
    end_time=datetime(2022, 5, 5),
    parameter_ids=PARAMETER_IDS,
    qualifier_ids=QUALIFIER_IDS,
    parallel=True,
)


def test_time_zone():
    assert time_series_set.time_zone == 1.0


def test_version():
    assert time_series_set.version == "1.31"


def test_empty():
    assert not time_series_set.empty


def test_length():
    assert len(time_series_set) == 2


def test_parameter_ids():
    assert all([i in PARAMETER_IDS for i in time_series_set.parameter_ids])


def test_location_ids():
    assert all([i in LOCATION_IDS for i in time_series_set.location_ids])


def test_qualifier_ids():
    assert time_series_set.qualifier_ids == ["productie"]



from fewspy.time_series import TimeSeriesSet
from pathlib import Path
import json


def read_json(json_path: Path) -> TimeSeriesSet:
    """Read the content of a JSON file into a fewspy TimeSeriesSet

    Args:
        json_path (Path): path to PI_JSON file

    Returns:
        TimeSeriesSet: timeseries
    """
    return TimeSeriesSet.from_dict(json.loads(Path(json_path).read_text()))



from .conversions import snake_to_camel_case, datetime_to_fews_str

DATETIME_KEYS = ["start_time", "end_time"]
API_KEYS = [
    "document_format",
    "document_version",
    "end_time",
    "filter_id",
    "include_location_relations",
    "location_ids",
    "only_headers",
    "omit_missing",
    "parameter_ids",
    "qualifier_ids",
    "attributes",
    "show_statistics",
    "start_time",
    "thinning",
]


def flatten_list(list_of_lists: list) -> list:
    """
    Flattens a list of lists

    Args:
        list_of_lists (list): List of lists

    Returns:
        list: flattened list

    """

    return [i for j in list_of_lists for i in j]


def parameters_to_fews(parameters: dict, bool_to_string=False) -> dict:
    """
    Prepare Python API dictionary for FEWS API request

    Args:
        parameters (dict): parameters passed by Python API

    Returns:
        dict: parameters prepard for FEWS API request

    """

    def _convert_kv(k: str, v) -> dict:
        if k in DATETIME_KEYS:
            v = datetime_to_fews_str(v)
        elif k == "attributes":
            k = "show_attributes"
            v = True
        k = snake_to_camel_case(k)

        if bool_to_string and (type(v) == bool):
            v = str(v)

        return k, v

    args = (_convert_kv(k, v) for k, v in parameters.items() if k in API_KEYS)
    args = (i for i in args if i[1] is not None)
    return {i[0]: i[1] for i in args}



import requests
import logging
import pandas as pd
from typing import List
from ..utils.timer import Timer
from ..utils.transformations import parameters_to_fews
from ..utils.conversions import camel_to_snake_case

LOGGER = logging.getLogger(__name__)
COLUMNS = [
    "id",
    "name",
    "parameter_type",
    "unit",
    "display_unit",
    "uses_datum",
    "parameter_group",
]


def get_parameters(
    url: str,
    filter_id: str = None,
    document_format: str = "PI_JSON",
    verify: bool = False,
    logger=LOGGER,
) -> List[dict]:
    """
    Get FEWS qualifiers as a pandas DataFrame

    Args:
        url (str): url Delft-FEWS PI REST WebService.
        E.g. http://localhost:8080/FewsWebServices/rest/fewspiservice/v1/qualifiers
        filter_id (str): the FEWS id of the filter to pass as request parameter
        document_format (str): request document format to return. Defaults to PI_JSON.
        verify (bool, optional): passed to requests.get verify parameter.
        Defaults to False.
        logger (logging.Logger, optional): Logger to pass logging to. By
        default, a logger will ge created.

    Returns:
        df (pandas.DataFrame): Pandas dataframe with index "id" and columns
        "name" and "group_id".

    """

    # do the request
    timer = Timer(logger)
    parameters = parameters_to_fews(locals())
    response = requests.get(url, parameters, verify=verify)
    timer.report("Parameters request")

    # parse the response
    df = pd.DataFrame(columns=COLUMNS)
    if response.status_code == 200:
        if "timeSeriesParameters" in response.json().keys():
            df = pd.DataFrame(response.json()["timeSeriesParameters"])
            df.columns = [camel_to_snake_case(i) for i in df.columns]
            df["uses_datum"] = df["uses_datum"] == "true"
            timer.report("Parameters parsed")
    else:
        logger.error(f"FEWS Server responds {response.text}")

    df.set_index("id", inplace=True)

    return df



import pandas as pd
from pathlib import Path
from config import api

DATA_PATH = Path(__file__).parent / "data"
PARAMETERS_CSV = DATA_PATH / "parameters.csv"
parameters_reference = pd.read_csv(PARAMETERS_CSV).set_index("id")
parameters = api.get_parameters()


def _write_reference_set():
    parameters.to_csv(PARAMETERS_CSV)


def test_to_refrence_set():
    assert parameters.sort_index().equals(parameters_reference.sort_index())



import requests
import pandas as pd
import geopandas as gpd
import logging
from ..utils.timer import Timer
from ..utils.transformations import parameters_to_fews
from ..utils.conversions import (
    attributes_to_array,
    camel_to_snake_case,
    geo_datum_to_crs,
    xy_array_to_point,
)


LOGGER = logging.getLogger(__name__)


def get_locations(
    url: str,
    filter_id: str = None,
    document_format: str = "PI_JSON",
    attributes: list = [],
    verify: bool = False,
    logger=LOGGER,
    remove_duplicates: bool = False
) -> pd.DataFrame:
    """
    Get FEWS qualifiers as a pandas DataFrame

    Args:
        url (str): url Delft-FEWS PI REST WebService.
        E.g. http://localhost:8080/FewsWebServices/rest/fewspiservice/v1/qualifiers
        filter_id (str): the FEWS id of the filter to pass as request parameter
        document_format (str): request document format to return. Defaults to PI_JSON.
        attributes (list): if not emtpy, the location attributes to include as columns in the pandas DataFrame.
        verify (bool, optional): passed to requests.get verify parameter.
        Defaults to False.
        logger (logging.Logger, optional): Logger to pass logging to. By default, a logger will ge created.

    Returns:
        df (pandas.DataFrame): Pandas dataframe with index "id" and columns
        "name" and "group_id".

    """

    # do the request
    timer = Timer(logger)
    parameters = parameters_to_fews(locals())
    response = requests.get(url, parameters, verify=verify)
    timer.report("Locations request")

    # parse the response
    if response.status_code == 200:
        # convert to gdf and snake_case
        gdf = gpd.GeoDataFrame(response.json()["locations"], geometry=gpd.GeoSeries())
        gdf.columns = [camel_to_snake_case(i) for i in gdf.columns]

        # remove duplicates
        if remove_duplicates:
            gdf.drop_duplicates(subset="location_id", inplace=True, ignore_index=True)

        # set index
        gdf.set_index("location_id", inplace=True)

        # handle geometry and crs
        gdf["geometry"] = xy_array_to_point(gdf[["x", "y"]].values)
        gdf.crs = geo_datum_to_crs(response.json()["geoDatum"])

        # handle attributes
        if attributes:
            gdf.loc[:, attributes] = attributes_to_array(
                gdf["attributes"].values, attributes
            )
        gdf.drop(columns=["attributes"], inplace=True)

        timer.report("Locations parsed")

    else:
        logger.error(f"FEWS Server responds {response.text}")
        gdf = gpd.GeoDataFrame()

    return gdf



from fewspy.cache.manifest import Manifest
from typing import Optional
import threading
import xarray as xr
import pandas as pd
from pathlib import Path
import numpy as np
from datetime import datetime


class TimeSeriesCache:
    def __init__(self, manifest: Manifest, last_manifest_mtime: Optional[float] = None):
        self.manifest = manifest
        self.last_manifest_mtime: float | None = last_manifest_mtime
        self._lock = threading.RLock()
        self._datasets: dict[str, xr.Dataset] = {}
        self._common_time_axis: pd.DatetimeIndex | None = None
        self._open_all()  #

    def _key_for(self, path: Path) -> str:
        return f"{path.parent.name}/{path.name}"

    def _get_open_ds(self, filter_id: str, parameter_id: str) -> xr.Dataset:
        entry = self.manifest.get_entry(filter_id=filter_id, parameter_id=parameter_id)
        key = self._key_for(entry.path)
        with self._lock:
            ds = self._datasets.get(key)
        if ds is None:
            # Fallback if not cached
            entry = self.manifest.get_entry(
                filter_id=filter_id, parameter_id=parameter_id
            )
            ds = xr.open_dataset(
                entry.path,
                engine="netcdf4",
                decode_times=True,
                mask_and_scale=True,
                cache=False,
            )
        return ds

    def _open_all(self):
        """Open all datasets and close old datasets"""
        new = {}
        for entry in self.manifest.files:
            key = self._key_for(entry.path)
            dataset = xr.open_dataset(
                entry.path,
                engine="netcdf4",
                decode_times=True,
                mask_and_scale=True,
                cache=False,
            )
            new[key] = dataset
        with self._lock:
            old = self._datasets
            self._datasets = new
        for dataset in old.values():
            try:
                dataset.close()
            except:
                pass

    @property
    def common_time_axis(self) -> pd.DatetimeIndex:
        if not self._datasets:
            raise ValueError(
                "No NetCDF DataSets open in cache. Check manifest-file and run self._open_all()"
            )

        if self._common_time_axis is None:
            # init idx_all
            idx_all: pd.DatetimeIndex | None = None

            # union all time-axis to idx_all
            for dataset in self._datasets.values():
                idx = pd.to_datetime(dataset["time"].values)

                if idx_all is None:
                    idx_all = idx
                else:
                    idx_all = idx_all.union(idx)

            if idx_all is not None:
                self._common_time_axis = idx_all.sort_values()
        return self._common_time_axis

    @classmethod
    def _decode_station_ids(cls, sid: xr.DataArray) -> np.ndarray:
        """
        Retourneert een 1D numpy array met dtype 'U' (unicode) zonder Python-loop.
        Werkt voor:
        - fixed-width bytes: dtype.kind == 'S'
        - unicode strings:   dtype.kind == 'U'
        - object arrays met mix van bytes/str: dtype.kind == 'O'
        """
        vals = sid.values

        # Case 1: fixed-width bytes (b'...') -> vectorized decode
        if vals.dtype.kind == "S":  # e.g. dtype('S20')
            return np.char.decode(vals, "utf-8", "replace")

        # Case 2: al unicode
        if vals.dtype.kind == "U":
            return vals

        # Case 3: object array met mix bytes/str
        if vals.dtype.kind == "O":
            obj = vals
            # mask voor bytes
            is_bytes = np.frompyfunc(lambda x: isinstance(x, (bytes, bytearray)), 1, 1)(
                obj
            ).astype(bool)

            out = np.empty(obj.shape, dtype=object)
            if is_bytes.any():
                # Alleen de bytes-subset in één keer decoden
                out[is_bytes] = np.char.decode(
                    obj[is_bytes].astype("S"), "utf-8", "replace"
                )
            if (~is_bytes).any():
                # Niet-bytes naar str (no-op voor str, netjes voor ints e.d.)
                out[~is_bytes] = obj[~is_bytes].astype(str)
            return out.astype("U")

        # Fallback: veilig naar unicode casten
        return vals.astype("U")

    @classmethod
    def from_manifest_file(cls, path: Path):
        manifest = Manifest.from_file(path)
        manifest.validate_files()
        return cls(manifest=manifest, last_manifest_mtime=path.stat().st_mtime)

    def refresh_if_changed(self, manifest_path: Path) -> bool:
        """Swap DataSets if manifest_json has changed

        Args:
            manifest_path (Path): Path to manifest.json

        Returns:
            bool: True if swapped, else False
        """
        mtime = manifest_path.stat().st_mtime
        if mtime == self.last_manifest_mtime:
            return False
        # read new manifest and validate
        new_manifest = Manifest.from_file(manifest_path)
        try:
            # mount new manifest file
            new_manifest.validate_files()
            self.manifest = new_manifest
            self.last_manifest_mtime = mtime

            # swap to new netcdfs and close old
            self._open_all()

            # reset derived arguments so they will be re-computed
            self._common_time_axis = None
            return True
        except ValueError:  # in case new manifest returns invalid files
            return False

    def get_time_series(
        self,
        filter_id: str,
        parameter_id: str,
        start_time: Optional[datetime | str] = None,
        end_time: Optional[datetime | str] = None,
        location_ids: Optional[list[str]] = None,
    ) -> pd.DataFrame:
        """fetch time series data from NetCDF file based on filter_id and parameter_id

        Args:
            filter_id (str): filter_id
            parameter_id (str): parameter_id
            start_time (Optional[datetime  |  str], optional): start_time Defaults to None.
            end_time (Optional[datetime  |  str], optional): end_time. Defaults to None.
            location_ids (Optional[list[str]], optional): location_ids. Defaults to None.

        Returns:
            pd.DataFrame: DataFrame with datetime index and MultiIndex columns (location_id, parameter_id)
        """
        dataset = self._get_open_ds(filter_id=filter_id, parameter_id=parameter_id)
        da = dataset[parameter_id]

        if "stations" in da.dims and "station_id" in dataset.variables:
            # swap station index for station_id values
            # TODO: bij chache een optie inbouwen om station_ids al als location_id (string) op te slaan
            # TODO: bij cache een optie inbouwen om station_ids te gebruiken i.p.v. stations index
            station_ids = self._decode_station_ids(dataset["station_id"])
            da = da.assign_coords(station_id=("stations", station_ids))
            da = da.swap_dims({"stations": "station_id"})

        # Slice time and location_ids
        slicer = {}
        if (start_time is not None) | (end_time is not None):
            slicer["time"] = slice(start_time, end_time)
        if location_ids is not None:
            slicer["station_id"] = location_ids
        if slicer:
            da = da.sel(**slicer)

        s = da.to_series()
        if "station_id" in s.index.names:
            df = s.unstack("station_id")
        else:
            df = s.unstack("stations")
            df.columns = df.columns.astype(str)

        # MultiIndex-kolommen (station_id, parameter_id)
        df.columns = pd.MultiIndex.from_arrays(
            [df.columns, [parameter_id] * len(df.columns)],
            names=["location_id", "parameter_id"],
        )

        # sort indices
        df.sort_index(inplace=True)
        df.index.name = "datetime"
        df.sort_index(inplace=True, axis=1)

        return df



import shutil
from datetime import datetime, timezone
from pathlib import Path

import numpy as np
import pandas as pd
from netCDF4 import Dataset, date2num


def _datetimeindex_to_nc_time(
    idx: pd.DatetimeIndex, units="seconds since 1970-01-01 00:00:00 UTC"
):
    # netCDF4.date2num expects naive datetimes + units/tz in string;
    py_dt = [
        d.to_pydatetime().replace(tzinfo=timezone.utc).replace(tzinfo=None) for d in idx
    ]
    return date2num(py_dt, units=units), units


def write_netcdf(
    df: pd.DataFrame,
    out_dir: Path,
    global_attributes: dict = {"source": "fewspy"},
    file_template: str = "{parameter_id}.nc",
    remove_dir: bool = False,
) -> None:
    """Write a pandas DataFrame to netCDF files, one per parameter_id.

    Args:
        df (pd.DataFrame): _description_
        out_dir (Path): _description_
        global_attributes (dict(str), optional): _description_. Defaults to {"source": "fewspy"}.
        file_template (str, optional): _description_. Defaults to "{parameter_id}.nc".
        remove_dir (bool, optional): If True, removes the output directory before writing. Defaults to False.
    """

    # prepare output directory
    if remove_dir:
        shutil.rmtree(out_dir, ignore_errors=True)
    out_dir.mkdir(exist_ok=True, parents=True)

    # write one netCDF file per parameter_id
    for parameter_id in set(df.columns.get_level_values(1)):
        # Filter dataframe for parameter_id and drop all-NaN rows
        dfp = df.loc[:, df.columns.get_level_values(1) == parameter_id].dropna(
            how="all"
        )

        # to numpy
        values = dfp.to_numpy(dtype=float)

        # prepare dimensions
        location_ids = dfp.columns.get_level_values(0).to_list()
        strlen = max(len(s) for s in location_ids)
        time_vals, time_units = _datetimeindex_to_nc_time(dfp.index)

        # create netCDF file
        nc_file = out_dir / file_template.format(parameter_id=parameter_id)
        with Dataset(nc_file, "w", format="NETCDF4_CLASSIC") as nc:
            n_time, n_stations = values.shape

            # dimensions
            nc.createDimension("time", n_time)
            nc.createDimension("stations", n_stations)
            nc.createDimension("char_leng_id", strlen)  # for station_id

            # variables: time
            vtime = nc.createVariable("time", "f8", ("time",))
            vtime.units = time_units
            vtime.standard_name = "time"
            vtime.long_name = "time"
            vtime.calendar = "gregorian"
            vtime[:] = time_vals

            # variables: station

            vstation = nc.createVariable(
                "station_id", "S1", ("stations", "char_leng_id")
            )
            vstation.long_name = "station identification code"
            vstation.cf_role = "timeseries_id"

            # convert list of strings to array of characters (FEWS style)
            arr = np.array(location_ids, dtype=f"S{strlen}")  # fixed-length bytestrings
            data = arr.view("S1").reshape(n_stations, strlen)
            vstation[:, :] = data

            # compression and chunks
            compression_args = dict(zlib=True, complevel=4, shuffle=True)
            chunks = (min(max(n_time // 10, 1), n_time), min(n_stations, 128))

            vval = nc.createVariable(
                parameter_id,
                "f4",
                ("time", "stations"),
                fill_value=np.nan,
                chunksizes=chunks,
                **compression_args,
            )
            vval.coordinates = "time station_id"
            vval[:] = values

            # Globale attributen (CF-vriendelijk)
            nc.Conventions = "CF-1.6"
            nc.featureType = "timeSeries"
            nc.history = f"Created {datetime.now(timezone.utc).isoformat()}Z"
            nc.parameter_id = parameter_id
            for key, value in global_attributes.items():
                setattr(nc, key, value)



# %%
from fewspy.time_series import TimeSeriesSet, Header, TimeSeries
from pathlib import Path
import pandas as pd
from fewspy.io.header_file import get_header_file


def _row_to_header(row):
    d = row.to_dict()
    header = d.copy()
    header["time_step"] = {
        "unit": header.pop("time_step.unit"),
        "multiplier": header.pop("time_step.multiplier"),
    }
    return Header(**header)


def _column_to_time_series(df, column):
    df = pd.DataFrame(df[column])
    df.columns = ["value"]
    return df


def read_parquet(parquet_file: Path) -> TimeSeriesSet:
    """Parse parquet file to fewspy TimeSeriesSet

    Args:
        parquet_file (Path): path to parquet-file

    Returns:
        TimeSeriesSet: timeseries
    """
    # header to list of dict
    header_df = pd.read_parquet(get_header_file(parquet_file))
    header_df.set_index(["location_id", "parameter_id"], drop=False, inplace=True)
    header_series = header_df.apply(_row_to_header, axis=1)

    # read timeseries
    df = pd.read_parquet(parquet_file, engine="pyarrow")

    time_series_set = TimeSeriesSet()

    time_series_set.time_series = [
        TimeSeries(
            header=i,
            events=_column_to_time_series(df, (i.location_id, i.parameter_id)),
        )
        for i in header_series
    ]

    return time_series_set



from fewspy import Api
from pathlib import Path

api = Api(
    url=r"https://www.hydrobase.nl/fews/nzv/FewsWebServices/rest/fewspiservice/v1/",
    ssl_verify=False,
)

data_dir = Path(__file__).parent / "data"



import datetime
import hashlib
import json
import os
import shutil
from pathlib import Path
from pydantic import BaseModel, field_validator
from fewspy import __version__ as fewspy_version
from fewspy.time_series import TimeSeriesSet


class FieldEndtry(BaseModel):
    path: Path
    nbytes: int
    sha256: str

    @property
    def name(self) -> str:
        return self.path.name

    @classmethod
    def _sha256(cls, path: Path) -> str:
        with path.open("rb") as f:
            return hashlib.file_digest(f, "sha256").hexdigest()

    @field_validator("nbytes")
    @classmethod
    def _non_negative(cls, v: int) -> int:
        if v < 0:
            raise ValueError("nbytes must be non-negative")
        return v

    @classmethod
    def from_file(cls, path: Path) -> "FieldEndtry":
        return cls(path=path, nbytes=path.stat().st_size, sha256=cls._sha256(path))


class Coverage(BaseModel):
    start_date: datetime.datetime | None = None
    end_date: datetime.datetime | None = None

    @field_validator("end_date")
    @classmethod
    def check_start_before_end(cls, v, values):
        start_date = values.data["start_date"]
        if start_date is not None and v <= start_date:
            raise ValueError("end_date must be after start_date")
        return v

    def update_coverage(
        self, start_date: datetime.datetime, end_date: datetime.datetime
    ):
        """Udate coverage start_date and end_date if the provided times are outside the current

        Parameters
        ----------
        start_date : datetime.datetime
           new start time to check
        end_date : datetime.datetime
            new end time to check
        """

        # update start_date if provided start_date is earlier or current is None
        if self.start_date is None:
            self.start_date = start_date
        elif start_date < self.start_date:
            self.start_date = start_date
        # update end_date if provided end_date is later or current is None
        if self.end_date is None:
            self.end_date = end_date
        elif end_date > self.end_date:
            self.end_date = end_date

    def update_coverage_from_timeserieset(self, tss: TimeSeriesSet):
        for ts in tss.time_series:
            self.update_coverage(
                start_date=ts.header.start_date, end_date=ts.header.end_date
            )


class Manifest(BaseModel):
    filepath: Path | None = None
    current_cache: str
    expected_file_count: int = 0
    max_cache_count: int = 3
    cache_dirs: list[Path] = []
    current_coverage: Coverage = Coverage()
    files: list[FieldEndtry] = []
    fewspy_version: str = fewspy_version

    @classmethod
    def current_cache_from_datetime(cls, current_cache_datetime) -> str:
        return current_cache_datetime.strftime("%Y%m%dT%H%M%S")

    @property
    def current_cache_datetime(self) -> datetime.datetime:
        return datetime.datetime.strptime(self.current_cache, "%Y%m%dT%H%M%S")

    @property
    def current_cache_dir(self) -> Path:
        if self.filepath is None:
            return None
        else:
            return self.filepath.parent.joinpath(self.current_cache)

    @classmethod
    def from_file(cls, filepath: Path, **kwargs) -> "Manifest":
        with open(filepath, "r", encoding="utf-8") as f:
            json_data = json.load(f)
            for k, v in kwargs.items():
                json_data[k] = v
            manifest = cls.model_validate(json_data)
            manifest.filepath = filepath
            manifest.update_filepaths(filepath)
            return manifest

    def update_filepaths(self, filepath: Path):
        """
        Update filepath, cache_dirs and files paths to a new filepath location.
        """
        # Update manifest filepath
        self.filepath = filepath

        # Update cache_dirs to new parent directory
        parent = filepath.parent
        self.cache_dirs = [parent.joinpath(d.name) for d in self.cache_dirs]

        # Update files paths to new parent directory
        for file_entry in self.files:
            filter_id, nc_file = file_entry.path.parts[-2:]
            file_entry.path = self.current_cache_dir.joinpath(filter_id, nc_file)

    def validate_files(self):
        """Validate file cache on expected number of files, filex existence, size and hash"""

        # Validate that the number of files matches the expected count
        if len(self.files) != self.expected_file_count:
            raise ValueError(
                f"Expected {self.expected_file_count} files, but got {len(self.files)}"
            )

        # Validate each file's existence, size, and hash
        errors = []
        for file_entry in self.files:
            # Check if file exists
            if not file_entry.path.exists():
                errors.append(f"File does not exist: {file_entry.path}")
                continue
            # Check file size
            actual_size = file_entry.path.stat().st_size
            if actual_size != file_entry.nbytes:
                errors.append(
                    f"Size mismatch for {file_entry.path}: expected {file_entry.nbytes}, got {actual_size}"
                )
                continue
            # Check file hash
            actual_hash = FieldEndtry._sha256(file_entry.path)
            if actual_hash != file_entry.sha256:
                errors.append(
                    f"Hash mismatch for {file_entry.path}: expected {file_entry.sha256}, got {actual_hash}"
                )
        if errors:
            raise ValueError("File validation errors:\n" + "\n".join(errors))

    def get_entry(self, filter_id: str, parameter_id: str) -> FieldEndtry | None:
        """Get a file entry by filter_id and location_id

        Parameters
        ----------
        filter_id : str
            The filter ID to search for
        parameter_id : str
            The parameter_id ID to search for

        Returns
        -------
        FieldEndtry | None
            The matching FieldEndtry if found. Otherwise, raises a ValueError.
        """
        entry = next(
            (
                i
                for i in self.files
                if (i.path.name == f"{parameter_id}.nc")
                and (i.path.parent.name == filter_id)
            ),
            None,
        )
        if entry is None:
            raise ValueError(
                f"No entry found for filter_id '{filter_id}' and parameter_id '{parameter_id}'"
            )

        return entry

    def clean_cache_dirs(self):
        """Clean all directories in root_dir except those in cache_dirs"""
        # Exclude current_cache from deletion

        current_cache_dir = self.current_cache_dir
        root_dir = current_cache_dir.parent
        remove_candidates = [
            d
            for d in root_dir.glob("*")
            if d.is_dir() and (d != current_cache_dir) and (d not in self.cache_dirs)
        ]  # make sure we don't do current cache
        for i in remove_candidates:
            if i is not self.current_cache_dir:  # extra check
                shutil.rmtree(i, ignore_errors=True)

    def update_cache_dirs(self):
        """Update cache_dirs list with current_cache_dir if not already present"""
        if (
            self.current_cache_dir is not None
            and self.current_cache_dir not in self.cache_dirs
        ):
            self.cache_dirs.append(self.current_cache_dir)

        self.cache_dirs = sorted(self.cache_dirs, reverse=True)[: self.max_cache_count]

    def atomic_write(self, filepath: Path, clean_old_caches: bool = True):
        """Atomic replacement of (an existing) manifest file

        Parameters
        ----------
        filepath : Path
            Path to (existing) manifest-file to write to
        clean_old_caches : bool, optional
            If True (default) clean old cache directories, by default True
        """

        # store filepath
        self.filepath = filepath
        tmp_filepath = filepath.with_name(f".{filepath.stem}.tmp.json")

        # Validate files before writing
        self.validate_files()

        # make sure we have a clean cache_dirs list
        self.update_cache_dirs()

        # atomic write data
        with open(tmp_filepath, "w", encoding="utf-8") as f:
            f.write(self.model_dump_json(indent=2))
            f.flush()  # Python to kernel
            os.fsync(f.fileno())  # Ensure data is written to disk
        os.replace(tmp_filepath, filepath)  # atomic replace

        # clean old cache dirs
        if clean_old_caches:
            self.clean_cache_dirs()



from datetime import datetime
from shapely.geometry import Point
import numpy as np
from typing import List

GEODATUM_MAPPING = {
    "WGS 1984": "epsg:4326",
    "Ordnance Survey Great Britain 1936": "epsg:4277",
    "TWD 1967": "epsg:3828",
    "Gauss Krueger Meridian2": None,
    "Gauss Krueger Meridian3": None,
    "Gauss Krueger Austria M34": "epsg:18009",
    "Gauss Krueger Austria M31": "epsg:18008",
    "Rijks Driehoekstelsel": "epsg:28992",
    "JRC": "epsg:3040",
    "DWD": None,
    "KNMI Radar": None,
    "CH1903": "epsg:4149",
    "PAK1": None,
    "PAK2": None,
    "SVY21": "epsg:3414",
}


def camel_to_snake_case(camel_case: str) -> str:
    """
    Convert camelCase to snake_case

    Args:
        camel_case (str): sentence in camelCase (e.g. myInputVariable)

    Returns:
        str: converted sentence in snake_case (in example my_input_variable)

    """

    return "".join(["_" + i.lower() if i.isupper() else i for i in camel_case]).lstrip(
        "_"
    )


def snake_to_camel_case(snake_case: str) -> str:
    """
    Convert snake_case to camelCase

    Args:
        snake_case (str): sentence in snake_case (in example my_input_variable)

    Returns:
        str: converted sentence in camelCase (e.g. myInputVariable)

    """

    words = snake_case.split("_")
    return words[0] + "".join(i.title() for i in words[1:])


def dict_to_datetime(data: dict) -> datetime:
    """
    Convert a FEWS PI datetime dict to datetime object

    Args:
        data (dict): FEWS PI datetime (e.g. {'date': '2022-05-01', 'time': '00:00:00'})

    Returns:
        datetime: Converted datetime object (in example datetime.datetime(2022, 5, 1, 0, 0))

    """

    if "time" in data.keys():
        time = data["time"]
    else:
        time = "00:00:00"

    date_time = datetime.fromisoformat(f'{data["date"]}T{time}')

    return date_time


def datetime_to_fews_str(date_time: datetime) -> str:
    """
    Convert a FEWS PI datetime dict to datetime object

    Args:
        date_time (datetime): datetime object (e.g. datetime.datetime(2022, 5, 1, 0, 0))


    Returns:
        datetime: Converted datetime for FEWS REST API format  %Y-%m-%dT%H:%M:%SZ
        (in example 2022-05-01T00:00:00Z)

    """

    return date_time.strftime("%Y-%m-%dT%H:%M:%SZ")


def xy_array_to_point(xy_array: np.ndarray) -> List[Point]:
    return [Point(i.astype(float)) for i in xy_array]


def attributes_to_array(attribute_values: np.ndarray, attributes: list) -> np.ndarray:
    def _get_values(x, attributes):
        selection = {i["id"]: i["value"] for i in x if i["id"] in attributes}
        return [selection[i] if i in selection.keys() else None for i in attributes]

    return np.array([_get_values(i, attributes) for i in attribute_values])


def geo_datum_to_crs(geo_datum: str) -> str:
    if geo_datum.startswith("UTM"):
        epsg_code = 32600
        zone = int(geo_datum[3:5].lstrip("0"))
        epsg_code += int(zone)
        if geo_datum[-1] == "S":
            epsg_code += 100
        crs = f"epsg:{epsg_code}"
    elif geo_datum.lower().startswith("epsg"):
        crs = geo_datum.lower()
    elif geo_datum in GEODATUM_MAPPING.keys():
        crs = GEODATUM_MAPPING[geo_datum]
    else:
        crs = None
    return crs



# %%
from fewspy.time_series import TimeSeriesSet
import pickle
from pathlib import Path


DATA_PATH = Path(__file__).parent / "data"
twinpy_tss = DATA_PATH.joinpath("io", "twin_fewspy_ts.pickle")

with open(twinpy_tss, "rb") as src:
    time_series_set: TimeSeriesSet = pickle.load(src)

out_dir = DATA_PATH.joinpath("io", "twinpy_to_netcdf")
global_attributes = {
    "institution": "HHNK",
    "source": "TWINpy",
    "title": "TWINpy export",
}

time_series_set.to_netcdf(out_dir=out_dir, global_attributes=global_attributes)



"""Timer for fewsbokeh."""

import time


class Timer(object):
    """Record function efficiency."""

    def __init__(self, logger):
        self.start = time.time()
        self.milestone = self.start
        self.logger = logger

    def start(self):
        """Start the timer."""
        self.start = time.time()

    def report(self, message=""):
        """Set milestone and report."""
        self.logger.debug(f"{message} in {(time.time() - self.milestone):.3f} sec")
        self.milestone = time.time()

    def reset(self, message=None):
        """Report task-efficiency and reset."""
        if message:
            self.logger.debug(f"{message} in {(time.time() - self.start):.3f} sec")
        self.start = time.time()
        self.milestone = self.start



# %%
import fewspy
from config import data_dir

EXPECTED_LOCATION_IDS = sorted(["CMB_03751-21", "CMB_6100-04"])
EXPECTED_PARAMETER_IDS = ["vullingsgraad"]
EXPECTED_QUALIFIER_IDS = []

XML_FILE = data_dir.joinpath("io", "sample.xml")
JSON_FILE = data_dir.joinpath("io", "sample.json")
NC_FILE = data_dir.joinpath("io", "sample.nc")

xml_ts = fewspy.read_xml(XML_FILE)
json_ts = fewspy.read_json(JSON_FILE)
nc_ts = fewspy.read_netcdf(NC_FILE)


def test_xml_ts():
    """Check xml time-series to expected values"""
    # check basic info
    assert xml_ts.version == "1.34"
    assert xml_ts.time_zone == 0.0

    # amount of time_series
    assert len(xml_ts) == 2

    # some checks in header
    assert sorted(xml_ts.location_ids) == EXPECTED_LOCATION_IDS
    assert xml_ts.parameter_ids == EXPECTED_PARAMETER_IDS
    assert xml_ts.qualifier_ids == EXPECTED_QUALIFIER_IDS

    # xcheck start_time and end_time
    expected_start_date = xml_ts.time_series[0].events.index[0]
    assert xml_ts.time_series[0].header.start_date == expected_start_date
    expected_end_date = xml_ts.time_series[0].events.index[-1]
    assert xml_ts.time_series[0].header.end_date == expected_end_date

    # xcheck interval
    expected_time_step = (
        (expected_end_date - expected_start_date) / (len(xml_ts.time_series[0]) - 1)
    ).total_seconds()

    assert xml_ts.time_series[0].header.time_step["multiplier"] == expected_time_step

    # check values by sum
    assert xml_ts.time_series[0].events["value"].sum() == 0
    assert xml_ts.time_series[1].events["value"].sum() == 6053.0


def test_json_ts():
    """Check json time-series to xml-timeseries


    With FEWS API a sightly different header with documentFormat=PI_XML and documentFormat=PI_JSON are returned
        - start_date in JSON is request start_data in XML first available data
        - z is not returned in XML, but is returned in JSON

    Therefore we pop start_date and z when comparing headers
    """
    # basic info
    assert json_ts.version == xml_ts.version
    assert json_ts.time_zone == xml_ts.time_zone

    # headers
    ignore_keys = ["start_date", "z"]

    json_header = {
        k: v
        for k, v in json_ts.time_series[0].header.__dict__.items()
        if k not in ignore_keys
    }
    xml_header = {
        k: v
        for k, v in xml_ts.time_series[0].header.__dict__.items()
        if k not in ignore_keys
    }

    assert json_header == xml_header

    json_header = {
        k: v
        for k, v in json_ts.time_series[1].header.__dict__.items()
        if k not in ignore_keys
    }
    xml_header = {
        k: v
        for k, v in xml_ts.time_series[1].header.__dict__.items()
        if k not in ignore_keys
    }

    assert json_header == xml_header

    # events
    assert json_ts.time_series[0].events.equals(xml_ts.time_series[0].events)
    assert json_ts.time_series[1].events.equals(xml_ts.time_series[1].events)


def test_netcdf_ts():
    """Check json time-series to xml-timeseries


    With NetCDF does not contain version-info. TimeZone will always be 0 (GMT) Per header is also doen't contain:
        - type
        - module_instance_id
    It does contain z (XML doesn't)

    Events do not contain flags.

    Therefore we don't compare these info
    """

    assert nc_ts.time_zone == 0.0
    # headers
    ignore_keys = ["type", "module_instance_id", "z"]

    nc_header = {
        k: v
        for k, v in nc_ts.time_series[0].header.__dict__.items()
        if k not in ignore_keys
    }

    xml_header = {
        k: v
        for k, v in xml_ts.time_series[0].header.__dict__.items()
        if k not in ignore_keys
    }

    assert nc_header == xml_header

    nc_header = {
        k: v
        for k, v in nc_ts.time_series[1].header.__dict__.items()
        if k not in ignore_keys
    }

    xml_header = {
        k: v
        for k, v in xml_ts.time_series[1].header.__dict__.items()
        if k not in ignore_keys
    }

    assert nc_header == xml_header

    # events
    assert (
        nc_ts.time_series[0]
        .events[["value"]]
        .equals(xml_ts.time_series[0].events[["value"]])
    )
    assert (
        nc_ts.time_series[1]
        .events[["value"]]
        .equals(xml_ts.time_series[1].events[["value"]])
    )


def test_parquet_ts(tmp_path):
    """Check json time-series to xml-timeseries


    With NetCDF does not contain version-info. TimeZone will always be 0 (GMT) Per header is also doen't contain:
        - type
        - module_instance_id
    It does contain z (XML doesn't)

    Events do not contain flags.

    Therefore we don't compare these info
    """

    # make sure parquet-file does not exist
    parquet_file = tmp_path.joinpath("io", "sample.parquet")
    parquet_file.unlink(missing_ok=True)
    assert not parquet_file.exists()

    # store xml_ts as parquet-file and check if exists
    xml_ts.to_parquet(parquet_file, include_header=True)
    assert parquet_file.exists()
    header_file = fewspy.io.header_file.get_header_file(parquet_file)
    assert header_file.exists()

    # read parquet time-series
    parquet_ts = fewspy.read_parquet(parquet_file=parquet_file)

    # headers
    ignore_keys = []

    parquet_header = {
        k: v
        for k, v in parquet_ts.time_series[0].header.__dict__.items()
        if k not in ignore_keys
    }

    xml_header = {
        k: v
        for k, v in xml_ts.time_series[0].header.__dict__.items()
        if k not in ignore_keys
    }

    assert parquet_header == xml_header

    parquet_header = {
        k: v
        for k, v in parquet_ts.time_series[1].header.__dict__.items()
        if k not in ignore_keys
    }

    xml_header = {
        k: v
        for k, v in xml_ts.time_series[1].header.__dict__.items()
        if k not in ignore_keys
    }

    assert parquet_header == xml_header

    # events
    assert (
        parquet_ts.time_series[0]
        .events[["value"]]
        .equals(xml_ts.time_series[0].events[["value"]])
    )
    assert (
        parquet_ts.time_series[1]
        .events[["value"]]
        .equals(xml_ts.time_series[1].events[["value"]])
    )



from pathlib import Path
import pandas as pd
from config import api


DATA_PATH = Path(__file__).parent / "data"
QUALIFIERS_CSV = DATA_PATH / "qualifiers.csv"
qualifiers_reference = pd.read_csv(QUALIFIERS_CSV).set_index("id")
qualifiers_reference.sort_index(inplace=True)


qualifiers = api.get_qualifiers()
qualifiers.sort_index(inplace=True)


def write_reference_set():
    qualifiers.to_csv(QUALIFIERS_CSV)


def test_to_refrence_set():
    all(qualifiers.index == qualifiers_reference.index)
    assert all(qualifiers.index == qualifiers_reference.index)



import requests


class URLNotFoundError(Exception):
    pass


def validate_url(url: str, test_postfix: str = "timezoneid") -> str:
    """

    Args:
        url: input url to be validated
        test_postfix: postfix to url used for testing. Defaults to 'filters'.

    Returns: validated url

    """

    # add / if not in input_url
    if not url.endswith("/"):
        url += "/"

    # test with request
    response = requests.get(f"{url}{test_postfix}", verify=False)
    if not response.ok:
        raise URLNotFoundError(f"{url} is not a root to a live FEWS PI Rest WebService")

    # estimate ssl_verify
    if url.startswith("https"):
        ssl_verify = True
    else:
        ssl_verify = False

    return url, ssl_verify



from pathlib import Path
import json
from config import api

DATA_PATH = Path(__file__).parent / "data"
FILTERS_JSON = DATA_PATH / "filters.json"
with open(FILTERS_JSON) as src:
    filter_reference = json.load(src)
filters = api.get_filters()


def _write_reference_set():
    with open(FILTERS_JSON, "w+") as dst:
        dst.write(json.dumps(filters))


def test_to_reference_set():
    assert filter_reference == filters



import requests
import pandas as pd
import logging
from fewspy.utils.timer import Timer
from fewspy.utils.transformations import parameters_to_fews
from typing import List, Union
from fewspy.time_series import TimeSeriesSet
from aiohttp import ClientSession

from datetime import datetime
import aiohttp
import asyncio
import nest_asyncio

nest_asyncio.apply()

LOGGER = logging.getLogger(__name__)


def __result_async_to_time_series_set(async_result):
    time_series_set = TimeSeriesSet()
    time_series_set_gen = (i for i in async_result if "timeSeries" if type(i) == dict)
    time_series_set_list = [i for i in time_series_set_gen if "timeSeries" in i.keys()]

    version = next((i for i in time_series_set_list if "version" in i.keys()), None)
    if version is not None:
        time_zone = next(
            (i for i in time_series_set_list if "timeZone" in i.keys()), None
        )
        if time_zone is not None:
            time_series = {
                "timeSeries": [
                    i["timeSeries"][0]
                    for i in time_series_set_list
                    if "timeSeries" in i.keys()
                ]
            }
            pi_time_series = {**version, **time_zone, **time_series}
            time_series_set = TimeSeriesSet.from_dict(pi_time_series)
    return time_series_set


def get_time_series_async(
    url: str,
    filter_id: str,
    location_ids: Union[str, List[str]] = None,
    parameter_ids: Union[str, List[str]] = None,
    qualifier_ids: Union[str, List[str]] = None,
    start_time: datetime = None,
    end_time: datetime = None,
    thinning: int = None,
    document_format: str = "PI_JSON",
    omit_missing: bool = True,
    verify: bool = False,
    logger=LOGGER,
) -> pd.DataFrame:
    """

    Args:
        url (str): url Delft-FEWS PI REST WebService.
        E.g. http://localhost:8080/FewsWebServices/rest/fewspiservice/v1/qualifiers
        filter_id (str): the FEWS id of the filter to pass as request parameter
        location_ids (list): list with FEWS location ids to extract timeseries from. Defaults to None.
        parameter_ids (list): list with FEWS parameter ids to extract timeseries from. Defaults to None.
        qualifier_ids (list): list with FEWS qualifier ids to extract timeseries from. Defaults to None.
        start_time (datetime.datetime): datetime-object with start datetime to use in request. Defaults to None.
        end_time (datetime.datetime): datetime-object with end datetime to use in request. Defaults to None.
        thinning (int): integer value for thinning parameter to use in request. Defaults to None.
        document_format (str): request document format to return. Defaults to PI_JSON.
        omit_missing (bool): if True, no missings values will be returned. Defaults to True
        verify (bool, optional): passed to requests.get verify parameter.
        Defaults to False.
        logger (logging.Logger, optional): Logger to pass logging to. By
        default, a logger will ge created.

    Returns:
        df (pandas.DataFrame): Pandas dataframe with index "id" and columns
        "name" and "group_id".

    """
    parameters = parameters_to_fews(locals(), bool_to_string=True)

    def _get_loop():
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
        finally:
            loop.set_debug(True)
            return loop

    async def get_timeseries_async(location_id, parameter_id, qualifier_id, session):
        """Get timerseries using FEWS (asynchronously)"""
        parameters["locationIds"] = [location_id]
        parameters["parameterIds"] = [parameter_id]
        if qualifier_id is not None:
            parameters["qualifierIds"] = qualifier_id
        try:
            response = await session.request(
                method="GET", url=url, params=parameters, ssl=verify
            )
            response.raise_for_status()
        except Exception as err:
            logger.error(
                f"An error ocurred: {err} while executing url {url} with parameters {parameters}"
            )
            response = None
        response_json = await response.json()
        return response_json

    async def run_program(location_id, parameter_id, qualifier_id, session):
        """Wrapper for running program in an asynchronous manner"""
        try:
            response = await get_timeseries_async(
                location_id, parameter_id, qualifier_id, session
            )
        except Exception as err:
            logger.error(f"Exception occured: {err}")
            response = None
            pass
        return response

    async def asynciee():
        async with aiohttp.ClientSession(loop=loop) as session:
            args = [
                (location_id, parameter_id)
                for location_id in location_ids
                for parameter_id in parameter_ids
            ]
            if qualifier_ids is None:
                args = [(*i, None) for i in args]
            else:
                args = [
                    (*i, qualifier_id) for i in args for qualifier_id in qualifier_ids
                ]
            fetch_all = [run_program(*i, session) for i in args]
            result_async = await asyncio.gather(*fetch_all)
            return result_async

    if __name__ == "fewspy.wrappers.get_time_series_async":
        loop = _get_loop()
        result_async = loop.run_until_complete(asynciee())
        time_series_set = __result_async_to_time_series_set(result_async)
    return time_series_set



from pathlib import Path
import geopandas as gpd
from config import api

ATTRIBUTES = ["MPN_IDENT", "MPN_BRON"]
DATA_PATH = Path(__file__).parent / "data"
REFERENCE_GPKG = DATA_PATH / "locations.gpkg"
locations_reference = gpd.read_file(REFERENCE_GPKG).set_index("location_id")
locations = api.get_locations(attributes=ATTRIBUTES)


def _write_reference_set():
    locations.to_file(REFERENCE_GPKG, driver="GPKG")


def test_to_reference_set():
    cols = [i for i in locations.columns if i != "geometry"]
    assert locations[cols].sort_index().equals(locations_reference[cols].sort_index())



import warnings
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import List, Literal, Optional, TypedDict

import pandas as pd

from fewspy.io.header_file import get_header_file
from fewspy.io.write_netcdf import write_netcdf
from fewspy.utils.conversions import camel_to_snake_case, dict_to_datetime
from fewspy.utils.transformations import flatten_list

DATETIME_KEYS = ["start_date", "end_date"]
FLOAT_KEYS = ["miss_val", "lat", "lon", "x", "y", "z"]
STRING_KEYS = ["module_instance_id"]
EVENT_COLUMNS = ["datetime", "value", "flag"]


class TimeStepDict(TypedDict, total=False):
    unit: Literal["second", "minute", "hour", "day", "month", "year", "nonequidistant"]
    multiplier: Optional[int]


def reliables(df: pd.DataFrame, threshold: int = 6) -> pd.DataFrame:
    """
    Filters reliables from an Events type Pandas DataFrame

    Args:
        df (pd.DataFrame): input Events-type Pandas Dataframe
        threshold (int, optional): threshold for unreleables. Defaults to 6.

    Returns:
        pd.DataFrame: Pandas DataFrame with reliable data only

    """
    if "flag" not in df.columns:
        return df
    else:
        return df.loc[df["flag"] < threshold]


@dataclass
class Header:
    """FEWS-PI header-style dataclass"""

    type: Literal["accumulative", "instantaneous"]
    module_instance_id: str | None
    location_id: str
    parameter_id: str
    time_step: TimeStepDict
    start_date: datetime
    end_date: datetime
    miss_val: float
    lat: float | None = None
    lon: float | None = None
    x: float | None = None
    y: float | None = None
    units: str | None = None
    station_name: str | None = None
    z: float | None = None
    qualifier_id: List[str] | None = None

    @classmethod
    def from_pi_header(cls, pi_header: dict):
        warnings.warn(
            "from_pi_header is depricated, use from_dict instead.", DeprecationWarning
        )
        return cls.from_dict(pi_header=pi_header)

    @classmethod
    def from_dict(cls, pi_header: dict):
        """
        Parse Header from FEWS PI header dict.

        Args:
            pi_header (dict): FEWS PI header as dictionary

        Returns:
            Header: FEWS-PI header-style dataclass

        """

        def _convert_kv(k: str, v) -> dict:
            k = camel_to_snake_case(k)
            if k in DATETIME_KEYS:
                v = dict_to_datetime(v)
            elif k in FLOAT_KEYS:
                v = float(v)
            elif k in STRING_KEYS:
                if v == "None":
                    v = None
                else:
                    str(v)
            elif k == "time_step":
                if "multiplier" in v.keys():
                    v["multiplier"] = float(v["multiplier"])
            return k, v

        args = (_convert_kv(k, v) for k, v in pi_header.items())
        return cls(**{i[0]: i[1] for i in args})

    def to_row(self):
        flat = self.__dict__.copy()
        ts = flat.pop("time_step", {})
        flat["time_step.unit"] = ts.get("unit")
        flat["time_step.multiplier"] = ts.get("multiplier")
        return flat


class Events(pd.DataFrame):
    """FEWS-PI events in pandas DataFrame"""

    @classmethod
    def from_pi_events(
        cls, pi_events: list, missing_value: float = None, tz_offset: float = None
    ):
        warnings.warn(
            "from_pi_events is deprecated, use from_dict instead.", DeprecationWarning
        )
        return cls.from_dict(
            pi_events=pi_events, missing_value=missing_value, tz_offset=tz_offset
        )

    @classmethod
    def from_dict(
        cls, pi_events: list, missing_value: float = None, tz_offset: float = None
    ):
        """
        Parse Events from FEWS PI events dict.

        Args:
            pi_events (dict): FEWS PI events as dictionary

        Returns:
            Events: pandas DataFrame

        """

        df = cls(pi_events)

        # set datetime
        if tz_offset is not None:
            df["datetime"] = pd.to_datetime(
                df["date"] + " " + df["time"]
            ) - pd.Timedelta(hours=tz_offset)
        else:
            df["datetime"] = pd.to_datetime(df["date"] + " " + df["time"])

        # drop columns and add missing columns
        drop_cols = [i for i in df.columns if i not in EVENT_COLUMNS]
        df.drop(columns=drop_cols, inplace=True)

        # set numeric types
        if "flag" not in df.columns:
            df["flag"] = pd.Series(dtype="int")
        else:
            df.loc[:, ["flag"]] = df["flag"].astype("int")

        df["value"] = pd.to_numeric(df["value"], downcast="float")

        # remove missings (if specified)
        if missing_value is not None:
            df = df.loc[df["value"] != missing_value]

        # set datetime to index
        df.set_index("datetime", inplace=True)

        return df


@dataclass
class TimeSeries:
    """FEWS-PI time series"""

    header: Header
    events: Events = field(
        default_factory=lambda: pd.DataFrame(columns=EVENT_COLUMNS).set_index(
            "datetime"
        )
    )

    def __len__(self):
        return len(self.events)

    @classmethod
    def from_pi_time_series(cls, pi_time_series: dict, time_zone: float = None):
        warnings.warn(
            "from_pi_time_series is deprecated, use from_dict instead.",
            DeprecationWarning,
        )
        return cls.from_dict(pi_time_series=pi_time_series, time_zone=time_zone)

    @classmethod
    def from_dict(cls, pi_time_series: dict, time_zone: float = None):
        """Parse TimeSeries from FEWS PI timeseries dict.

        Args:
            pi_time_series (dict): FEWS PI timeseries as dictionary
            time_zone (float, optional): time_zone. Defaults to None.

        Returns:
            fewspy.TimeSeries: time series in FEWS PI format
        """
        header = Header.from_dict(pi_time_series["header"])
        kwargs = dict(header=header)
        if "events" in pi_time_series.keys():
            kwargs["events"] = Events.from_dict(
                pi_time_series["events"], header.miss_val, time_zone
            )
        return cls(**kwargs)


@dataclass
class TimeSeriesSet:
    """FEWS-PI time series set"""

    version: str = None
    time_zone: float = None
    time_series: List[TimeSeries] = field(default_factory=list)

    def __len__(self):
        return len(self.time_series)

    @classmethod
    def from_pi_time_series(cls, pi_time_series_set):
        warnings.warn(
            "from_pi_time_series is deprecated, use from_dict instead.",
            DeprecationWarning,
        )
        return cls.from_dict(pi_time_series_set)

    @classmethod
    def from_dict(cls, pi_time_series_set: dict):
        """Parse TimeSeries from FEWS PI time series set dict.

        Args:
            pi_time_series_set (dict): FEWS PI time series set as dictionary

        Returns:
            fewspy.TimeSeriesSet: Time series set with multiple time series
        """
        kwargs = {}
        if "version" in pi_time_series_set.keys():
            kwargs["version"] = pi_time_series_set["version"]
        if "timeZone" in pi_time_series_set.keys():
            time_zone = float(pi_time_series_set["timeZone"])
            kwargs["time_zone"] = time_zone
        else:
            time_zone = None
        if "timeSeries" in pi_time_series_set.keys():
            kwargs["time_series"] = [
                TimeSeries.from_dict(i, time_zone)
                for i in pi_time_series_set["timeSeries"]
            ]
        return cls(**kwargs)

    def add(self, time_series_set):
        # add time_series to the time_series_set
        self.time_series += [time_series_set]
        return self

    @property
    def empty(self):
        return all([i.events.empty for i in self.time_series])

    @property
    def parameter_ids(self):
        return list(set([i.header.parameter_id for i in self.time_series]))

    @property
    def location_ids(self):
        return list(set([i.header.location_id for i in self.time_series]))

    @property
    def qualifier_ids(self):
        qualifiers = (i.header.qualifier_id for i in self.time_series)
        qualifiers = [i for i in qualifiers if i is not None]

        return list(set(flatten_list(qualifiers)))

    def to_df(self) -> pd.DataFrame:
        columns = pd.MultiIndex.from_tuples(
            [(i.header.location_id, i.header.parameter_id) for i in self.time_series],
            names=["location_id", "parameter_id"],
        )
        df = pd.concat([reliables(i.events)["value"] for i in self.time_series], axis=1)
        df.columns = columns

        return df

    def to_netcdf(
        self,
        out_dir: Path,
        global_attributes: dict = {"source": "fewspy"},
        file_template: str = "{parameter_id}.nc",
        remove_dir: bool = False,
    ) -> None:
        """Write fewspy.TimeSeriesSet to netCDF files, one per parameter_id.

        Args:
            out_dir (Path): Directory to save NetCDF files.
            global_attributes (dict, optional): Global attributes for the NetCDF files. Defaults to {"source": "fewspy"}.
            file_template (str, optional): Template for naming the NetCDF files. Defaults to "{parameter_id}.nc".
            remove_dir (bool, optional): If True, removes the output directory before writing. Defaults to False.
        """
        if not self.empty:
            df = self.to_df()

            write_netcdf(
                df=df,
                out_dir=out_dir,
                global_attributes=global_attributes,
                file_template=file_template,
                remove_dir=remove_dir,
            )

    def to_parquet(self, parquet_file: Path, include_header: bool = False):
        """Write fewspy.TimeSeriesSet to arrow parquet file

        Args:
            parquet_file (Path): parquet-file to store
            include_header (bool, optional): if true all headers will be stored as a parquet-file next to the timeseries. Defaults to False.
        """

        # make dir-structure to file(s)
        parquet_file.parent.mkdir(exist_ok=True, parents=True)
        parquet_file.unlink(missing_ok=True)

        # concat events to one dataframe and write to parquet
        df = self.to_df()
        df.to_parquet(parquet_file, engine="pyarrow")

        # if include_header, write header_file
        if include_header:
            header_file = get_header_file(parquet_file)
            header_file.unlink(missing_ok=True)
            header_df = pd.DataFrame([i.header.to_row() for i in self.time_series])
            header_df.to_parquet(header_file, engine="pyarrow")



from fewspy.api import Api
from fewspy.io.read_xml import read_xml
from fewspy.io.read_json import read_json
from fewspy.io.read_netcdf import read_netcdf
from fewspy.io.read_parquet import read_parquet
from fewspy.io.write_netcdf import write_netcdf
from fewspy.time_series import TimeSeries, TimeSeriesSet

__all__ = ["Api", "read_xml", "read_json", "read_netcdf", "read_parquet", "write_netcdf", "TimeSeries", "TimeSeriesSet"]
__version__ = "2025.11.0"


from datetime import datetime
from config import api

kwargs = dict(
    filter_id="WDB_OW_KGM",
    location_ids=["NL34.HL.KGM156.HWZ1", "NL34.HL.KGM156.LWZ1"],
    start_time=datetime(2022, 5, 1),
    end_time=datetime(2022, 5, 5),
    parameter_ids=["Q [m3/s] [NVT] [OW]", "WATHTE [m] [NAP] [OW]"],
)

timeseriesset = api.get_time_series(**kwargs)


def test_time_zone():
    assert timeseriesset.time_zone == 1.0


def test_version():
    assert timeseriesset.version == "1.31"


def test_empty():
    assert not timeseriesset.empty


def test_length():
    assert len(timeseriesset) == 2


def test_parameter_ids():
    assert timeseriesset.parameter_ids == ["WATHTE [m] [NAP] [OW]"]


def test_location_ids():
    assert all(
        [
            i in ["NL34.HL.KGM156.LWZ1", "NL34.HL.KGM156.HWZ1"]
            for i in timeseriesset.location_ids
        ]
    )


def test_qualifier_ids():
    assert timeseriesset.qualifier_ids == ["productie"]



#%%
from pathlib import Path
import pickle
import os

from TWINpy.app.TWIN_get_timeseries import TWIN_Timeserie


data_dir = Path(__file__).parent / "data"
os.environ["TWIN_BASE_URL"] = ""

with open(data_dir / "twin_response.pickle", "rb") as src:
    b = pickle.load(src)
# df = pd.read_pickle(data_dir / "twin_response.pickle")


#%%
from fewspy.time_series import TimeSeries, TimeSeriesSet, Header, Events

def convert_timeseries(time_series:TWIN_Timeserie, header_info: dict = {"type": "instantaneous", "time_step": {"unit": "nonequidistant"}}) -> TimeSeries:
    """Converts TWIN_Timeserie to fewspy.TimeSeriesSet

    Parameters
    ----------
    time_series : TWIN_Timeserie
    header_info : dict, optional
        Header-info missing in TWIN_Timeserie, by default {"type": "instantaneous", "time_step": {"unit": "nonequidistant"}}

    Returns
    -------
    TimeSeries
        _description_
    """
    header = {**time_series.ts_header, **header_info}
    header["miss_val"] = header.pop("missing_val")
    header["start_date"] = time_series.df.index.min()
    header["end_date"] = time_series.df.index.min()

    return TimeSeries(header=Header(**header), events=Events(time_series.df))
 
time_series_set = TimeSeriesSet(time_series = [convert_timeseries(i) for i in b.values()])
time_series_set.to_netcdf(out_dir=Path("netcdf"))
# %%



from fewspy.wrappers.get_time_series_async import get_time_series_async
from fewspy.wrappers.get_qualifiers import get_qualifiers
from fewspy.wrappers.get_time_series import get_time_series
from fewspy.wrappers.get_locations import get_locations
from fewspy.wrappers.get_filters import get_filters
from fewspy.wrappers.get_parameters import get_parameters
from fewspy.wrappers.get_timezone_id import get_timezone_id

__all__ = ["get_time_series_async", "get_qualifiers", "get_time_series", "get_locations", "get_filters", "get_parameters", "get_timezone_id"]


# %%
from lxml import etree
from pathlib import Path
from fewspy.utils.conversions import snake_to_camel_case
from fewspy.time_series import TimeSeriesSet
from io import BytesIO

ns = {"pi": "http://www.wldelft.nl/fews/PI"}


def _parse_xml_data(xml_data) -> TimeSeriesSet:
    root = xml_data.getroot()  # Root element

    version = root.attrib.get("version")
    time_zone_element = root.find("pi:timeZone", namespaces=ns)
    time_zone = float(time_zone_element.text) if time_zone_element is not None else None

    time_series_set = {"version": version, "timeZone": time_zone, "timeSeries": []}

    # Get children, filter out timezone.
    rootchildren = [
        child for child in root.getchildren() if child.tag.endswith("series")
    ]

    # Loop over children (individual timeseries in the xml)
    for child in rootchildren:
        subchildren = child.getchildren()  # alle headers en en timevalues
        metadata = None
        data = []

        for subchild in subchildren:
            # Write header to dict
            if subchild.tag.endswith(
                "header"
            ):  # gewoonlijk eerste subchild is de header
                header_childs = subchild.getchildren()
                metadata = {}
                for item in header_childs:
                    key = item.tag.split("}")[-1]

                    item_keys = item.keys()

                    if len(item_keys) == 0:
                        metadata[key] = item.text
                    else:
                        metadata[key] = {}
                        for item_key, item_value in zip(item_keys, item.values()):
                            metadata[key][item_key] = item_value
            # Get event data
            else:
                data += [{k: v for k, v in zip(subchild.keys(), subchild.values())}]

        # add to time_series_set
        if "start_date" not in metadata:
            metadata["start_date"] = {"date": data[0]["date"], "time": data[0]["time"]}
        if "end_date" not in metadata:
            metadata["end_date"] = {"date": data[-1]["date"], "time": data[-1]["time"]}

        time_series_set["timeSeries"] += [{"header": metadata, "events": data}]

    return TimeSeriesSet.from_dict(time_series_set)


def read_xml(xml_path: Path) -> TimeSeriesSet:
    """Parse PI XML file to fewspy TimeSeriesSet

    Args:
        xml_path (Path): Path to xml-file

    Returns:
        TimeSeriesSet: timeseries
    """

    xml_data = etree.parse(xml_path)  # Parse XML data
    return _parse_xml_data(xml_data)


def read_xml_from_string(xml_string: str) -> TimeSeriesSet:
    """Parse PI XML file to fewspy TimeSeriesSet

    Args:
        xml_string (str): string with PI_XML data

    Returns:
        TimeSeriesSet: timeseries
    """

    xml_data = etree.fromstring(xml_string.encode("utf-8"))
    return _parse_xml_data(xml_data)



"""
Module for calling the FEWS REST API.

The module contains one class and methods corresponding with the FEWS PI-REST requests:
https://publicwiki.deltares.nl/display/FEWSDOC/FEWS+PI+REST+Web+Service
"""

import pandas as pd
from .utils.timer import Timer
from .utils.url import validate_url
import logging
import urllib3

from fewspy.wrappers import (
    get_time_series_async,
    get_qualifiers,
    get_time_series,
    get_locations,
    get_filters,
    get_parameters,
    get_timezone_id,
)

LOGGER = logging.getLogger(__name__)

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


class Api:
    """
    Python API for the Deltares FEWS PI REST Web Service.

    For more info on how-to work with the FEWS REST Web Service, visit the Deltares Website: https://publicwiki.deltares.nl/display/FEWSDOC/FEWS+PI+REST+Web+Service
    """

    def __init__(self, url, logger=None, ssl_verify=None):
        self.document_format = "PI_JSON"
        self.logger = logger
        self.timer = Timer(logger)
        self.url, verify = validate_url(url)

        # set ssl_verify
        if ssl_verify is None:
            self.ssl_verify = verify
        else:
            self.ssl_verify = ssl_verify

        # set logger
        if logger is None:
            self.logger = LOGGER
        else:
            self.logger = logger

    def __kwargs(self, url_post_fix: str, kwargs: dict) -> dict:
        kwargs = {
            **kwargs,
            **dict(
                url=f"{self.url}{url_post_fix}",
                verify=self.ssl_verify,
                logger=self.logger,
            ),
        }
        kwargs.pop("self")
        kwargs.pop("parallel", None)
        return kwargs

    def get_parameters(self, filter_id=None):
        """
        Get FEWS qualifiers as a pandas DataFrame

        Args:
            filter_id (str): the FEWS id of the filter to pass as request parameter

        Returns:
            df (pandas.DataFrame): Pandas dataframe with index "id" and columns
            "name" and "group_id".

        """

        kwargs = self.__kwargs(url_post_fix="parameters", kwargs=locals())
        result = get_parameters(**kwargs)

        return result

    def get_filters(self, filter_id=None):
        """
        Get FEWS qualifiers as a pandas DataFrame

        Args:
            E.g. http://localhost:8080/FewsWebServices/rest/fewspiservice/v1/qualifiers
            filter_id (str): the FEWS id of the filter to pass as request parameter

        Returns:
            df (pandas.DataFrame): Pandas dataframe with index "id" and columns
            "name" and "group_id".

        """

        kwargs = self.__kwargs(url_post_fix="filters", kwargs=locals())
        result = get_filters(**kwargs)

        return result

    def get_locations(self, filter_id=None, attributes=[], remove_duplicates=False):
        """
        Get FEWS qualifiers as a pandas DataFrame

        Args:
            E.g. http://localhost:8080/FewsWebServices/rest/fewspiservice/v1/qualifiers
            filter_id (str): the FEWS id of the filter to pass as request parameter
            attributes (list): if not emtpy, the location attributes to include as columns in the pandas DataFrame.
            remove_duplicates (bool): if True, duplicated location_ids are removed. Default = False

        Returns:
            df (pandas.DataFrame): Pandas dataframe with index "id" and columns
            "name" and "group_id".

        """

        kwargs = self.__kwargs(url_post_fix="locations", kwargs=locals())
        result = get_locations(**kwargs)

        return result

    def get_qualifiers(self) -> pd.DataFrame:
        """
        Get FEWS qualifiers as Pandas DataFrame

        Returns:
            df (pandas.DataFrame): Pandas dataframe with index "id" and columns
            "name" and "group_id".

        """
        url = f"{self.url}qualifiers"
        result = get_qualifiers(url, verify=self.ssl_verify, logger=self.logger)
        return result

    def get_timezone_id(self):
        """
        Get FEWS timezone_id

        Returns:
            str: timezone id FEWS API is running on

        """
        url = f"{self.url}timezoneid"
        result = get_timezone_id(url, verify=self.ssl_verify, logger=self.logger)
        return result

    def get_time_series(
        self,
        filter_id,
        location_ids=None,
        start_time=None,
        end_time=None,
        parameter_ids=None,
        qualifier_ids=None,
        thinning=None,
        only_headers=False,
        omit_missing=True,
        show_statistics=False,
        parallel=False,
        document_format:str="PI_JSON"
    ):
        """
        Get FEWS qualifiers as a pandas DataFrame

        Args:
            filter_id (str): the FEWS id of the filter to pass as request parameter
            location_ids (list): list with FEWS location ids to extract timeseries from. Defaults to None.
            parameter_ids (list): list with FEWS parameter ids to extract timeseries from. Defaults to None.
            qualifier_ids (list): list with FEWS qualifier ids to extract timeseries from. Defaults to None.
            start_time (datetime.datetime): datetime-object with start datetime to use in request. Defaults to None.
            end_time (datetime.datetime): datetime-object with end datetime to use in request. Defaults to None.
            thinning (int): integer value for thinning parameter to use in request. Defaults to None.
            only_headers (bool): if True, only headers will be returned. Defaults to False.
            omit_missing (bool): if True, no missings values will be returned. Defaults to True.
            show_statistics (bool): if True, time series statistics will be included in header. Defaults to False.
            document_format (str): request document format to return. Defaults to PI_JSON.
            parallel (bool): if True, timeseries are requested by the asynchronous wrapper. Defaults to False

        Returns:
            df (pandas.DataFrame): Pandas dataframe with index "id" and columns
            "name" and "group_id".

        """
        kwargs = self.__kwargs(url_post_fix="timeseries", kwargs=locals())
        if (self.document_format != "PI_JSON") and parallel:
            self.logger.warning("Wont run parallel, as this is only supported for documentFromat PI_JSON") 
            parallel = False
        if parallel:
            kwargs.pop("only_headers")
            kwargs.pop("show_statistics")
            result = get_time_series_async(**kwargs)
        else:
            result = get_time_series(**kwargs)

        return result



from pathlib import Path
import json
from fewspy.time_series import TimeSeriesSet

DATA_PATH = Path(__file__).parent / "data"

with open(DATA_PATH / "pi_time_series.json") as src:
    pi_time_series = json.load(src)

timeseriesset = TimeSeriesSet.from_dict(pi_time_series)


def test_time_zone():
    assert timeseriesset.time_zone == 1.0


def test_version():
    assert timeseriesset.version == "1.28"


def test_empty():
    assert not timeseriesset.empty


def test_length():
    assert len(timeseriesset) == 2


def test_parameter_ids():
    assert timeseriesset.parameter_ids == ["WATHTE [m] [NAP] [OW]"]


def test_location_ids():
    assert all(
        [
            i in ["NL34.HL.KGM156.LWZ1", "NL34.HL.KGM156.HWZ1"]
            for i in timeseriesset.location_ids
        ]
    )


def test_qualifier_ids():
    assert timeseriesset.qualifier_ids == ["validatie"]



import requests
import logging
from ..utils.timer import Timer
from ..utils.transformations import parameters_to_fews

LOGGER = logging.getLogger(__name__)


def get_timezone_id(
    url: str,
    filter_id: str = None,
    document_format: str = "PI_JSON",
    verify: bool = False,
    logger=LOGGER,
) -> str:
    """
    Get FEWS timezone id

    Args:
        url (str): url Delft-FEWS PI REST WebService.
        E.g. http://localhost:8080/FewsWebServices/rest/fewspiservice/v1/timezoneid
        filter_id (str): the FEWS id of the filter to pass as request parameter
        document_format (str): request document format to return. Defaults to PI_JSON.
        verify (bool, optional): passed to requests.get verify parameter.
        Defaults to False.
        logger (logging.Logger, optional): Logger to pass logging to. By
        default, a logger will ge created.

    Returns:
        str: timezone string, e.g. GMT+01:00 expressing a GMT + 1 hour offset

    """

    # do the request
    timer = Timer(logger)
    parameters = parameters_to_fews(locals())
    response = requests.get(url, parameters, verify=verify)
    timer.report("Timezone request")

    # parse the response
    if response.status_code == 200:
        result = response.text
        timer.report("Timezone parsed")
    else:
        logger.error(f"FEWS Server responds {response.text}")

    return result



from netCDF4 import Dataset, num2date
from pathlib import Path
import pandas as pd
from fewspy.time_series import TimeSeriesSet, TimeSeries, Header
import zipfile
from io import BytesIO
import tempfile
import os


type = "instantaneous"
module_instance_id = "Vullingsgraad"


def _parse_time(time_var):
    times = num2date(time_var[:], units=time_var.units, only_use_cftime_datetimes=False)
    time_index = pd.to_datetime(times)
    time_index.name = "datetime"
    return time_index


def _get_time_step(time_index):
    # time delta
    deltas = time_index.to_series().diff().dropna()
    first_delta = deltas.iloc[0]
    equidistant = (deltas == first_delta).all()
    if equidistant:
        return {"unit": "second", "multiplier": first_delta.total_seconds()}
    else:
        return {"unit": "nonequidistant"}


def _parse_locations(stations_var):
    station_id_var = stations_var[:]

    return [
        "".join([c.decode("utf-8") if isinstance(c, bytes) else "" for c in row])
        .strip()
        .replace("\x00", "")
        for row in station_id_var
    ]


def _get_parameter_id(ds):
    parameter_ids = [
        var_name
        for var_name, var in ds.variables.items()
        if var.dimensions == ("time", "stations")
    ]
    return parameter_ids


def read_netcdf_from_content(content) -> TimeSeriesSet:
    """Read zipped NetCDF content as TimeSeriesSet."""
    with zipfile.ZipFile(BytesIO(content)) as zf:
        nc_file_name = next(
            (name for name in zf.namelist() if name.endswith(".nc")), None
        )
        if nc_file_name is None:
            raise ValueError(
                f"No NetCDF-file in content, with filelist {zf.namelist()}"
            )

        # Create a temp file path without opening the file
        fd, tmp_path = tempfile.mkstemp(suffix=".nc")
        os.close(fd)  # Close the low-level file descriptor

        try:
            with open(tmp_path, "wb") as f:
                f.write(zf.read(nc_file_name))  # write zip contents to temp file

            result = read_netcdf(Path(tmp_path))
        finally:
            os.remove(tmp_path)
        return result


def read_netcdf(
    nc_file: Path,
    time_series_type: str | None = None,
    module_instance_id: str | None = None,
) -> TimeSeriesSet:
    """Read the content of a NetCDF file into a fewspy TimeSeriesSet

    Args:
        nc_file (Path): path to the NetCDF file
        time_series_type (str | None, optional): type for timeseries header. Defaults to None.
        module_instance_id (str | None, optional): ModuleInstanceId for timeseries header. Defaults to None.

    Returns:
        TimeSeriesSet: timeseries
    """
    # Read file
    with Dataset(nc_file, mode="r") as ds:
        # init TimeSeriesSet
        time_series_set = TimeSeriesSet(time_zone=0.0)

        # Get time-index for events
        time_index = _parse_time(ds.variables["time"])
        time_step = _get_time_step(time_index)
        start_date = time_index[0].to_pydatetime()
        end_date = time_index[-1].to_pydatetime()

        # Get Locations
        location_ids = _parse_locations(ds.variables["station_id"])
        location_names = _parse_locations(ds.variables["station_names"])

        # Get coordinates
        x = list(ds.variables["x"][:].data)
        y = list(ds.variables["y"][:].data)
        lat = list(ds.variables["lat"][:].data)
        lon = list(ds.variables["lon"][:].data)
        z_fill = ds.variables["z"]._FillValue
        z = [None if i == z_fill else float(i) for i in ds.variables["z"][:].data]

        # Get Parameters
        parameter_ids = _get_parameter_id(ds)

        # Populate TimeSeries
        for parameter_id in parameter_ids:
            var = ds.variables[parameter_id]
            miss_val = float(var._FillValue)
            for i in range(len(location_ids)):
                # define header
                header = Header(
                    type=time_series_type,
                    module_instance_id=module_instance_id,
                    location_id=location_ids[i],
                    parameter_id=parameter_id,
                    time_step=time_step,
                    start_date=start_date,
                    end_date=end_date,
                    x=float(x[i]),
                    y=float(y[i]),
                    lat=float(lat[i]),
                    lon=float(lon[i]),
                    units=var.units,
                    station_name=location_names[i],
                    z=z[i],
                    qualifier_id=None,
                    miss_val=miss_val,
                )

                # define events
                data = {"value": pd.to_numeric(var[:, i].data, downcast="float")}
                events = pd.DataFrame(data=data, index=time_index)

                # append to TimeSeriesSet
                time_series_set.time_series.append(
                    TimeSeries(header=header, events=events)
                )

    return time_series_set
